/*  load.c

   for S2 cpu simulator
   5 December 2001
   modify from s1 sim 1997, 1998

   supporting function

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

	add print, printc  to output buffer  31 Dec 2001
	add MOS support						 1 Apr 2016
*/

#include "s21.h"

//int itime[] = {
//6,6,4,6,6,4,5,		/* lda ldd ldi sta std jmp jal */
//5,5,5,5,5,5,5, 		/* addi subi muli divi andi ori xori */
//1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 17 undef */
//5,5,5,5,5,5,5,5,5,	/* add sub mul div and or xor shl shr */
//6,6,4,4  			/* ldx stx jr trap */
//};

int clock, ninst;
FILE *fp;
int ip;

extern int R[32], Ir, Pc, M[], RetAds;
extern int runflag, savePc, ninst;

PRIVATE char ob[256];	// output buffer

void error(char *s){
	printf("error: %s at %d\n",s,Pc);
	exit(0);
}
/*
void countclock(){
  int i;
  if( IRop() == 31 )
    i = 31 + IRxop();
  else
  	i = IRop();
  clock += itime[i];
  ninst++;
}

void dump(int ads, int n){	// dump n words starting from ads
  int i;
  for(i=ads; i<ads+n; i++)
    printf("%4d %d\n",i,M[i]);
}

void report(void){
  float cpi;
  if( ninst == 0 )
	  printf("0 instruction\n");
  else {
  	  cpi = ((float) clock)/ninst;
  	  printf("%d instructions, %d clocks, CPI %4.2f\n",ninst,clock,cpi);
  }
}
*/
PRIVATE int signx( int d ){		// sign bit 16 extended
	if( d & 0x010000 ) return d | 0xFFFE0000;
	return d;
}

// IR bits
int IRop(void)  { return((Ir & 0xF8000000) >> 27);} // bit 31..27
int IRr1(void)  { return((Ir & 0x07C00000) >> 22);} // bit 26..22
int IRads(void) { return( Ir & 0x003FFFFF       );} // bit 21..0
int IRr2(void)  { return((Ir & 0x003E0000) >> 17);} // bit 21..17
int IRr3(void)  { return((Ir & 0x0001F000) >> 12);} // bit 16..12
int IRxop(void) { return( Ir & 0x00000FFF       );} // bit 11..0
int IRdisp(void){ return( signx(Ir & 0x0001FFFF));} // bit 16..0

extern char *cp2;

// -----load object code from file -----

// get one charactor skip newline and blank
PRIVATE char geta(void){
  char s[32];
  fscanf(fp,"%s",s);
  return s[0] ;
}

PRIVATE int geti(void){	// get one integer
  int c;
  fscanf(fp,"%d",&c);
  return c;
}

// set maximum of number to be read to 20000
//   to prevent infinite loop when input is incorrect
void loadprogram(char *name){
  int cnt, a1, a2, a3, a4, a5;

  fp = fopen(name,"r");
  if(fp == NULL){
	 printf("input file not found");
	 exit(0);
  }
  ip = 0;
  for(cnt=0; cnt<20000; cnt++ ) {
    switch( geta() ) {
    case 'a': ip = geti(); break;			// set address
    case 'L':
		a1 = geti(); a2 = geti(); a3 = geti();
		M[ip] = a1<<27 | a2<<22 | (a3&0x3FFFFF);
		ip++;
		break;
	case 'D':
		a1 = geti(); a2 = geti(); a3 = geti(); a4 = geti();
		M[ip] = a1<<27 | a2<<22 | a3<<17 | (a4&0x1FFFF);
		ip++;
		break;
	case 'X':
		a1=geti(); a2=geti(); a3=geti(); a4=geti(); a5=geti();
		M[ip] = a1<<27 | a2<<22 | a3<<17 | a4<<12 | (a5&0x0FFF);
		ip++;
		break;
    case 'w': M[ip] = geti(); ip++; break;	// set constant data
    case 'e': cnt = 200001;	break;			// end
	default: error("incorrect object file");
    }
  }
  fclose(fp);
  printf("load program, last address %d\n",ip);
}

PRIVATE void pr3R(char *s){
	sprintf(ob,"%s r%d r%d r%d",s,IRr1(),IRr2(),IRr3());
}
PRIVATE void prI(char *s) {
	sprintf(ob,"%s r%d r%d #%d",s,IRr1(),IRr2(),IRdisp());
}
PRIVATE void prA(char *s) {
	sprintf(ob,"%s r%d %d",s,IRr1(),IRads());
}

PRIVATE void pr2R(char *s){
	sprintf(ob,"%s r%d r%d",s,IRr1(),IRr2());
}

PRIVATE void pr1R(char *s){
	sprintf(ob,"%s r%d",s,IRr1());
}

PRIVATE void disassem(void){
	switch( IRop() ) {
	case LDA: prA("ld"); break;
	case LDD: sprintf(ob,"ld r%d @%d r%d",IRr1(),IRdisp(),IRr2()); break;
	case STA: prA("st"); break;
	case STD: sprintf(ob,"st r%d @%d r%d",IRr1(),IRdisp(),IRr2()); break;
	case JMP: sprintf(ob,"jmp %d",IRads()); break;
	case JAL: prA("jal"); break;
	case JT:  prA("jt"); break;
	case JF:  prA("jf"); break;
	case MVI: sprintf(ob,"mv r%d #%d",IRr1(),signx2(IRads())); break;
	case ADDI:prI("add"); break;
	case SUBI:prI("sub"); break;
	case MULI:prI("mul"); break;
	case DIVI:prI("div"); break;
	case ANDI:prI("and"); break;
	case ORI: prI("or"); break;
	case XORI:prI("xor"); break;
	case EQI: prI("eq"); break;
	case NEI: prI("ne"); break;
	case LTI: prI("lt"); break;
	case LEI: prI("le"); break;
	case GTI: prI("gt"); break;
	case GEI: prI("ge"); break;
	case SHLI:prI("shl"); break;
	case SHRI:prI("shr"); break;
	case XOP:
		switch(IRxop()){
		case ADD:pr3R("add"); break;
		case SUB:pr3R("sub"); break;
		case MUL:pr3R("mul"); break;
		case DIV:pr3R("div"); break;
		case AND:pr3R("and"); break;
		case OR: pr3R("or"); break;
		case XOR:pr3R("xor"); break;
		case EQ: pr3R("eq"); break;
		case NE: pr3R("ne"); break;
		case LT: pr3R("lt"); break;
		case LE: pr3R("le"); break;
		case GT: pr3R("gt"); break;
		case GE: pr3R("ge"); break;
		case SHL:pr3R("shl"); break;
		case SHR:pr3R("shr"); break;
		case MV: pr2R("mv"); break;
		case PUSH: pr2R("push"); break;
		case POP: pr2R("pop"); break;
		case LDX:sprintf(ob,"ld r%d +r%d r%d",IRr1(),IRr2(),IRr3()); break;
		case STX:sprintf(ob,"st r%d +r%d r%d",IRr1(),IRr2(),IRr3()); break;
		case RET: pr1R("ret"); break;
		case TRAP: pr1R("trap"); break;
		case NOT: pr2R("not"); break;
		case INT: pr1R("int"); break;
		case RETI:sprintf(ob,"reti"); break;
		case SAVR: pr1R("savr"); break;
		case RESR: pr1R("resr"); break;
		case SAVT: pr1R("savt"); break;
		case REST: pr1R("rest"); break;
		default: sprintf(ob,"unknown op code");
		}
		break;
	default: sprintf(ob,"unknown op code");
	}
}

extern int Q[], qindex;

void show(void){		// show some registers
  int i;
  printf("PC%4d ",savePc);
  disassem();
  for(i=strlen(ob); i<18; i++)	// pad blank to 18
	  ob[i] = 32;
  ob[i] = 0;
  printf("%s",ob);
  for(i=1;i<10;i++)
	  printf("r%d:%d ",i,R[i]);
  printf("\n");
  for(i=20;i<30;i++)
	  printf("r%d:%d ",i,R[i]);
  printf("\n");
//  printf("r27:%d r29:%d r31:%d\n",R[27],R[29],R[31]);
//  printf("Retads %d #inst %d\n",RetAds,ninst);

  printf("q ix %d :",qindex);
  for(i = 0; i < 16; i++)
	  printf("%d ",Q[i]);
  printf("\n");
  printf("PCB [%d %d] [%d %d]\n",M[2100],M[2101],M[2116],M[2117]);
}
